import type {
  Document,
  DirEntry,
  FileEntry,
  FUID,
  DirList,
  SelectedItems
} from '@/repositories/Document'
import { formatSize, formatUnixDate, haystackFormat } from '@/utils'
import { defineStore } from 'pinia'
import { collator } from '@/utils'

type FileData = { id: string; mtime: number; size: number; dir: DirectoryData }
type DirectoryData = {
  [filename: string]: FileData
}
type User = {
  username: string
  privileged: boolean
  isOpenLoginModal: boolean
  isLoggedIn: boolean
}

export const useDocumentStore = defineStore({
  id: 'documents',
  state: () => ({
    root: {} as DirEntry,
    document: [] as Document[],
    search: "" as string,
    selected: new Set<FUID>(),
    uploadingDocuments: [],
    uploadCount: 0 as number,
    wsWatch: undefined,
    wsUpload: undefined,
    fileExplorer: null,
    error: '' as string,
    user: {
      username: '',
      privileged: false,
      isLoggedIn: false,
      isOpenLoginModal: false
    } as User
  }),

  actions: {
    updateRoot(root: DirEntry | null = null) {
      root ??= this.root
      // Transform tree data to flat documents array
      let loc = ""
      const mapper = ([name, attr]: [string, FileEntry | DirEntry]) => ({
        loc,
        name,
        type: 'dir' in attr ? 'folder' : 'file' as 'folder' | 'file',
        ...attr,
        sizedisp: formatSize(attr.size),
        modified: formatUnixDate(attr.mtime),
        haystack: haystackFormat(name),
      })
      const queue = [...Object.entries(root.dir ?? {}).map(mapper)]
      const docs = []
      for (let doc; (doc = queue.shift()) !== undefined;) {
        docs.push(doc)
        if ("dir" in doc) {
          loc = doc.loc ? `${doc.loc}/${doc.name}` : doc.name
          queue.push(...Object.entries(doc.dir).map(mapper))
        }
      }
      // Pre sort directory entries folders first then files, names in natural ordering
      docs.sort((a, b) =>
        // @ts-ignore
        (a.type === "file") - (b.type === "file") ||
        collator.compare(a.name, b.name)
      )
      this.root = root
      this.document = docs
    },
    updateUploadingDocuments(key: number, progress: number) {
      for (const d of this.uploadingDocuments) {
        if (d.key === key) d.progress = progress
      }
    },
    pushUploadingDocuments(name: string) {
      this.uploadCount++
      const document = {
        key: this.uploadCount,
        name: name,
        progress: 0
      }
      this.uploadingDocuments.push(document)
      return document
    },
    deleteUploadingDocument(key: number) {
      this.uploadingDocuments = this.uploadingDocuments.filter(e => e.key !== key)
    },
    updateModified() {
      for (const d of this.document) {
        if ('mtime' in d) d.modified = formatUnixDate(d.mtime)
      }
    },
    login(username: string, privileged: boolean) {
      this.user.username = username
      this.user.privileged = privileged
      this.user.isLoggedIn = true
      this.user.isOpenLoginModal = false
    }
  },
  getters: {
    isUserLogged(): boolean {
      return this.user.isLoggedIn
    },
    recentDocuments(): Document[] {
      const ret = [...this.document]
      ret.sort((a, b) => b.mtime - a.mtime)
      return ret
    },
    largeDocuments(): Document[] {
      const ret = [...this.document]
      ret.sort((a, b) => b.size - a.size)
      return ret
    },
    selectedFiles(): SelectedItems {
      function traverseDir(data: DirEntry | FileEntry, path: string, relpath: string) {
        if (!('dir' in data)) return
        for (const [name, attr] of Object.entries(data.dir)) {
          const fullname = path ? `${path}/${name}` : name
          const key = attr.key
          // Is this the file we are looking for? Ignore if nested within another selection.
          let r = relpath
          if (selected.has(key) && !relpath) {
            ret.selected.add(key)
            ret.rootdir[name] = attr
            r = name
          } else if (relpath) {
            r = `${relpath}/${name}`
          }
          if (r) {
            ret.entries[key] = attr
            ret.fullpath[key] = fullname
            ret.relpath[key] = r
            ret.ids.push(key)
            if (!('dir' in attr)) ret.url[key] = `/files/${fullname}`
          }
          traverseDir(attr, fullname, r)
        }
      }
      const selected = this.selected
      const ret: SelectedItems = {
        selected: new Set<FUID>(),
        missing: new Set<FUID>(),
        rootdir: {} as DirList,
        entries: {} as Record<FUID, FileEntry | DirEntry>,
        fullpath: {} as Record<FUID, string>,
        relpath: {} as Record<FUID, string>,
        url: {} as Record<FUID, string>,
        ids: [] as FUID[]
      }
      traverseDir(this.root, '', '')
      // What did we not select?
      for (const id of selected) {
        if (!ret.selected.has(id)) ret.missing.add(id)
      }
      // Sorted array of FUIDs for easy traversal
      ret.ids.sort((a, b) =>
        ret.relpath[a].localeCompare(ret.relpath[b], undefined, {
          numeric: true,
          sensitivity: 'base'
        })
      )
      return ret
    }
  }
})
